-------------------------------------------------------------------------------
-- Title      : controlUnit
-- Project    :
-------------------------------------------------------------------------------
-- File       : operativeUnit.vhd
-- Author     : Jean-Noel BAZIN  <jnbazin@pc-disi-026.enst-bretagne.fr>
-- Company    :
-- Created    : 2018-04-11
-- Last update: 2019-02-13
-- Platform   :
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: Control unit of a sequential FIR filter.
-------------------------------------------------------------------------------
-- Copyright (c) 2018
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2018-04-11  1.0      jnbazin Created
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity controlUnit is

  port (
    I_clock               : in  std_logic;  -- global clock
    I_reset               : in  std_logic;  -- asynchronous global reset
    I_inputSampleValid    : in  std_logic;  -- Control signal to load the input sample in the sample shift register and shift the register
    I_processingDone      : in  std_logic;
    O_loadShift_X         : out std_logic;  -- filtered sample
    O_loadShift_Y         : out std_logic;  -- filtered sample
    O_loadShift_Z         : out std_logic;  -- filtered sample
    O_initAddress         : out std_logic;  -- Control signal to initialize register read address
    O_incrAddress         : out std_logic;  -- Control signal to increment register read address
    O_initSum             : out std_logic;  -- Control signal to initialize the MAC register
    O_loadSum             : out std_logic;  -- Control signal to load the MAC register;
    O_load_Output         : out std_logic;  -- Control signal to load Y register
    O_select              : out std_logic_vector(1 downto 0);
    O_FilteredSampleValid : out std_logic  -- Data valid signal for filtered sample
    );

end entity controlUnit;
architecture archi_operativeUnit of controlUnit is


  type T_state is (WAIT_SAMPLE, STORE_X, PROCESSING_LOOP_FIR1, STORE_Y, PROCESSING_LOOP_IIR_Y,
  RESET_ADDRESS_Y, PROCESSING_LOOP_IIR_Z, STORE_Z, PROCESSING_LOOP_FIR2, OUTPUT, WAIT_END_SAMPLE);  -- state list
  signal SR_presentState : T_state;
  signal SR_futurState   : T_state;

begin

  process (I_clock, I_reset) is
  begin
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_presentState <= WAIT_SAMPLE;
    elsif rising_edge(I_clock) then     -- rising clock edge
      SR_presentState <= SR_futurState;
    end if;
  end process;

  process (I_clock) is
  begin
    case SR_presentState is

      when WAIT_SAMPLE =>
        if (I_inputsampleValid = '1') then
          SR_futurState <= STORE_X;
        end if;

      when STORE_X =>
        SR_futurState <= PROCESSING_LOOP_FIR1;

      when PROCESSING_LOOP_FIR1 =>
        if (I_processingDone = '1') then
          SR_futurState <= STORE_Y;
          SR_futurState <= STORE_Y;
        end if;

      when STORE_Y =>
        SR_futurState <= PROCESSING_LOOP_IIR_Y;

      when PROCESSING_LOOP_IIR_Y =>
        if (I_processingDone = '1') then
          SR_futurState <= RESET_ADDRESS_Y;
        end if;

      when RESET_ADDRESS_Y =>
        SR_futurState <= PROCESSING_LOOP_IIR_Z;

      when PROCESSING_LOOP_IIR_Z =>
        if (I_processingDone = '1') then
          SR_futurState <= STORE_Z;
        end if;

      when STORE_Z =>
        SR_futurState <= PROCESSING_LOOP_FIR2;

      when PROCESSING_LOOP_FIR2 =>
        if (I_processingDone = '1') then
          SR_futurState <= OUTPUT;
        end if;

      when OUTPUT =>
      
        SR_futurState <= WAIT_END_SAMPLE;      
      
      when others =>
        if (I_inputSampleValid = '0') then
          SR_futurState <= WAIT_SAMPLE;
        end if;

      
    end case;
  end process;

  O_loadShift_X           <= '1' when (SR_presentState = STORE_X) else '0';
  O_loadShift_Y           <= '1' when (SR_presentState = STORE_Y) else '0';
  O_loadShift_Z           <= '1' when (SR_presentState = STORE_Z) else '0';
  O_initAddress           <= '1' when  (SR_presentState = STORE_X or SR_presentState = STORE_Y or SR_presentState = RESET_ADDRESS_Y or SR_presentState = STORE_Z) else '0';
  O_incrAddress           <= '1' when (SR_presentState = PROCESSING_LOOP_FIR1 or SR_presentState = PROCESSING_LOOP_IIR_Y or SR_presentState = PROCESSING_LOOP_IIR_Z or SR_presentState = PROCESSING_LOOP_FIR2) ;
  O_initSum               <= '1' when  (SR_presentState = STORE_X or SR_presentState = STORE_Y or SR_presentState = STORE_Z) else '0';
  O_loadSum               <= '1' when (SR_presentState = PROCESSING_LOOP_FIR1 or SR_presentState = PROCESSING_LOOP_IIR_Y or SR_presentState = PROCESSING_LOOP_IIR_Z or SR_presentState = PROCESSING_LOOP_FIR2) ;
  O_load_Output           <= '1' when (SR_presentState = OUTPUT) else '0' ;
  O_select                <= "01" when (SR_presentState = STORE_Y or SR_presentState = PROCESSING_LOOP_IIR_Y) else
                            "10" when (SR_presentState = RESET_ADDRESS_Y or SR_presentState = PROCESSING_LOOP_IIR_Z) else 
                            "11" when (SR_presentState = STORE_Z or SR_presentState = PROCESSING_LOOP_FIR2) else
                            "00"; 
  O_FilteredSampleValid <= '1' when (SR_presentState = OUTPUT) else '0' ;





end architecture archi_operativeUnit;
