-------------------------------------------------------------------------------
-- Title      : operativeUnit
-- Project    :
-------------------------------------------------------------------------------
-- File       : operativeUnit.vhd
-- Author     : Jean-Noel BAZIN  <jnbazin@pc-disi-026.enst-bretagne.fr>
-- Company    :
-- Created    : 2018-04-11
-- Last update: 2019-02-13
-- Platform   :
-- Standard   : VHDL'93/02
-------------------------------------------------------------------------------
-- Description: Operative unit of a sequential FIR filter. Including shift
-- register for samples, registers for coefficients, a MAC and a register to
-- store the result
-------------------------------------------------------------------------------
-- Copyright (c) 2018
-------------------------------------------------------------------------------
-- Revisions  :
-- Date        Version  Author  Description
-- 2019-02-13  1.1      marzel  Update to provide a 16-tap filter and improve
--                              the user experience ;)
-- 2018-04-11  1.0      jnbazin Created
-- 2018-04-18  1.0      marzel  Modification of SR_Y assignment to a round
--                              instead of a trunc
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity operativeUnit is

  port (
    I_clock          : in  std_logic;   -- global clock
    I_reset          : in  std_logic;   -- asynchronous global reset
    I_inputSample    : in  std_logic_vector(10 downto 0);  -- 8 bit input sample
    I_loadShift_X      : in  std_logic;  -- Control signal to load the input sample in the sample shift register and shift the register
    I_loadShift_Y      : in  std_logic;  -- Control signal to load the input sample in the sample shift register and shift the register
    I_loadShift_Z      : in  std_logic;  -- Control signal to load the input sample in the sample shift register and shift the register
    I_initAddress    : in  std_logic;  -- Control signal to initialize register read address
    I_incrAddress    : in  std_logic;  -- Control signal to increment register read address
    I_initSum        : in  std_logic;  -- Control signal to initialize the MAC register
    I_loadSum        : in  std_logic;  -- Control signal to load the MAC register;
    I_loadY          : in  std_logic;   -- Control signal to load Y register
    O_processingDone : out std_logic;   -- Indicate that processing is done
    O_Y              : out std_logic_vector(10 downto 0);   -- filtered sample
    I_select         : in std_logic_vector(1 downto 0)   
    );

end entity operativeUnit;

architecture arch_operativeUnit of operativeUnit is
  type registerFile94 is array(0 to 94) of signed(10 downto 0);     --Regfile of 11 bits for data
  type registerFile2 is array(0 to 2) of signed(10 downto 0);
  type registerFile10 is array(0 to 10) of signed(10 downto 0);

  type registerFile47 is array(0 to 47) of signed(11 downto 0);     --Regfile of 12 bits for coef
  type registerFile1 is array(0 to 1) of signed(11 downto 0);
  type registerFile5 is array(0 to 5) of signed(11 downto 0);

  
  signal SR_coefRegister_0 : registerFile47;
  signal SR_coefRegister_1 : registerFile1;
  signal SR_coefRegister_2 : registerFile1;
  signal SR_coefRegister_3 : registerFile5;


  signal SR_shiftRegister_X : registerFile94;  -- shift register file used to store and shift input samples
  signal SR_shiftRegister_Y : registerFile2;  -- shift register file used to store and shift input samples
  signal SR_shiftRegister_Z : registerFile10;  -- shift register file used to store and shift input samples
  signal SC_multOperand1  : signed(10 downto 0);
  signal SC_multOperand2  : signed(11 downto 0);
  signal SC_MultResult    : signed(22 downto 0);  -- Result of the multiplication Xi*Hi
  signal SC_addResult     : signed(28 downto 0);  -- result of the accumulation addition
  signal SR_sum           : signed(28 downto 0);  -- Accumulation register
  signal SR_Y             : signed(10 downto 0);  -- filtered sample storage register
  signal SR_readAddress   : integer range 0 to 94;  -- register files read address
  signal SR_readAddress_0   : integer range 0 to 94;  -- register files read address
  signal SR_readAddress_1   : integer range 0 to 2;  -- register files read address
  signal SR_readAddress_2   : integer range 0 to 1;  -- register files read address
  signal SR_readAddress_3   : integer range 0 to 10;  -- register files read address
  signal SR_Address_max     : integer range 0 to 94;


begin

-- Low-pass filter provided with octave (or Matlab ;)) command
--fir1(15, .001)/sqrt(sum(fir1(15, .001).^2))*2^6
  SR_coefRegister_0 <= (to_signed(-1, 12),
                        to_signed(-1, 12),
                        to_signed(-1, 12),
                        to_signed(-1, 12),
                        to_signed(-1, 12),
                        to_signed(-1, 12),
                        to_signed(-2, 12),
                        to_signed(-2, 12),
                        to_signed(-2, 12),
                        to_signed(-3, 12),
                        to_signed(-3, 12),
                        to_signed(-3, 12),
                        to_signed(-4, 12),
                        to_signed(-4, 12),
                        to_signed(-5, 12),
                        to_signed(-5, 12),
                        to_signed(-6, 12),
                        to_signed(-6, 12),
                        to_signed(-7, 12),
                        to_signed(-7, 12),
                        to_signed(-8, 12),
                        to_signed(-8, 12),
                        to_signed(-9, 12),
                        to_signed(-10, 12),
                        to_signed(-10, 12),
                        to_signed(-11, 12),
                        to_signed(-11, 12),
                        to_signed(-12, 12),
                        to_signed(-13, 12),
                        to_signed(-13, 12),
                        to_signed(-14, 12),
                        to_signed(-14, 12),
                        to_signed(-15, 12),
                        to_signed(-15, 12),
                        to_signed(-16, 12),
                        to_signed(-16, 12),
                        to_signed(-17, 12),
                        to_signed(-17, 12),
                        to_signed(-18, 12),
                        to_signed(-18, 12),
                        to_signed(-18, 12),
                        to_signed(-19, 12),
                        to_signed(-19, 12),
                        to_signed(-19, 12),
                        to_signed(-19, 12),
                        to_signed(-19, 12),
                        to_signed(-19, 12),
                        to_signed(1004, 12)
                      );
                      
   SR_coefRegister_1 <= (to_signed(480, 12),
                      to_signed(-777, 12)    
                      );
   SR_coefRegister_2 <= (to_signed(-777, 12),
                      to_signed(449, 12)
                      );
   SR_coefRegister_3 <= (to_signed(-129, 12),
                        to_signed(122, 12),
                        to_signed(149, 12),
                        to_signed(191, 12),
                        to_signed(226, 12),
                        to_signed(239, 12)
                      );

    shift_X : process (I_clock, I_reset) is
  begin  -- process shift
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_shiftRegister_X <= (others => (others => '0'));
    elsif rising_edge(I_clock) then
      if I_loadShift_X = '1' then
        -- Décalage à droite des valeurs dans le registre
        for i in 94 downto 1 loop
          SR_shiftRegister_X(i) <= SR_shiftRegister_X(i-1);
        end loop;
        -- Chargement du nouvel échantillon en première position
        SR_shiftRegister_X(0) <= signed(I_inputSample);
      end if;
    end if;
  end process shift_X;
  
  
    shift_Y : process (I_clock, I_reset) is
  begin  -- process shift
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_shiftRegister_Y <= (others => (others => '0'));
    elsif rising_edge(I_clock) then
      if I_loadShift_Y = '1' then
        -- Décalage à droite des valeurs dans le registre
        for i in 2 downto 1 loop
          SR_shiftRegister_Y(i) <= SR_shiftRegister_Y(i-1);
        end loop;
        -- Chargement du nouvel échantillon en première position
        SR_shiftRegister_Y(0) <= to_signed(to_integer(SR_sum(28 downto 10)),11);
      end if;
    end if;
  end process shift_Y;
  
  
    shift_Z : process (I_clock, I_reset) is
  begin  -- process shift
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_shiftRegister_Z <= (others => (others => '0'));
    elsif rising_edge(I_clock) then
      if I_loadShift_Z = '1' then
        -- Décalage à droite des valeurs dans le registre
        for i in 10 downto 1 loop
          SR_shiftRegister_Z(i) <= SR_shiftRegister_Z(i-1);
        end loop;
        -- Chargement du nouvel échantillon en première position
        SR_shiftRegister_Z(0) <= to_signed(to_integer(SR_sum(28 downto 10)),11);
      end if;
    end if;
  end process shift_Z;

SR_readAddress_0 <= SR_readAddress when I_select = "00" else 0 ;
SR_readAddress_1 <= SR_readAddress mod 4 when I_select = "01" else 0 ;
SR_readAddress_2 <= SR_readAddress mod 2 when I_select = "10" else 0 ;
SR_readAddress_3 <= SR_readAddress mod 16 when I_select = "11" else 0 ;

SR_Address_max <= 94 when (I_select = "00")
             else 2 when (I_select = "01")
             else 1 when (I_select = "10")
             else 10;

  incr_address : process (I_clock, I_reset) is
  begin
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_readAddress <= 0;
    elsif rising_edge(I_clock) then
      if I_initAddress = '1' then
        SR_readAddress <= 0;
      elsif I_incrAddress = '1' and SR_readAddress < SR_Address_max then
        SR_readAddress <= SR_readAddress+1;
      else
        SR_readAddress <= SR_readAddress;
      end if;
      else 
      SR_readAddress <= SR_readAddress;
    end if;
  end process incr_address;

  O_processingDone <= '1' when SR_readAddress = SR_Address_max else '0' ;

  SC_multOperand1 <= SR_shiftRegister_X(SR_readAddress_0/2) when I_select = "00" and SR_readAddress_0 mod 2 = 0 else
                     SR_shiftRegister_X(94 - SR_readAddress_0/2) when I_select = "00" and SR_readAddress_0 mod 2 = 1 else
                     SR_shiftRegister_Y(SR_readAddress_1/2) when I_select = "01" and SR_readAddress_1 mod 2 = 0 else
                     SR_shiftRegister_Y(2 - SR_readAddress_1/2) when I_select = "01" and SR_readAddress_1 mod 2 = 1 else
                     SR_shiftRegister_Z(SR_readAddress_2) when I_select = "10" else
                     SR_shiftRegister_Z(SR_readAddress_3/2) when I_select = "11" and SR_readAddress_3 mod 2 = 0 else
                     SR_shiftRegister_Z(10 - SR_readAddress_3/2);-- when I_select = "11" and SR_readAddress_3 mod 2 = 1 else
                     --0
                     
  SC_multOperand2 <= SR_coefRegister_0(SR_readAddress_0/2) when I_select = "00" else
                     SR_coefRegister_1(SR_readAddress_1/2) when I_select = "01" else
                     SR_coefRegister_2(SR_readAddress_2) when I_select = "01" else
                     SR_coefRegister_3(SR_readAddress_3/2);
  
  
  SC_MultResult   <= SC_multOperand1 * SC_multOperand2 ;  -- 16 bits
  SC_addResult    <= resize(SC_MultResult, SC_addResult'length) + SR_sum;

  sum_acc : process (I_clock, I_reset) is
  begin
    if I_reset = '1' then               -- asynchronous reset (active high)
      SR_sum <= (others => '0');
    elsif rising_edge(I_clock) then
      if I_initSum = '1' then
        SR_sum <= (others => '0');
      elsif I_loadSum = '1' then
        SR_sum <= SC_addResult;
      else
        SR_sum <= SR_sum;
      end if;
    end if;
  end process sum_acc;

  store_result : process (I_clock, I_reset) is
  begin
      if I_reset = '1' then               -- asynchronous reset (active high)
        SR_Y <= (others => '0');
    elsif rising_edge(I_clock) then
        if (I_loadY = '1') then
            
          SR_Y <= to_signed(to_integer(SR_sum(28 downto 10)),11);
          --SR_Y <= SR_sum(20 downto 10);
         
        end if;
      end if;

  end process store_result;

  O_Y <= std_logic_vector(SR_Y);

end architecture arch_operativeUnit;
